/* 
 * Module Name:     aiptouch.c
 *
 * Description:     IBM Touch Calibration RS-232 Read Monitor Driver for Linux
 *
 *  (C) Copyright 2005 International Business Machines Corporation.
 *  All right reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1.  Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 * 2.  Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * 3.  The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <gha@us.ibm.com>.
 */


/* System Headers */
#include <linux/version.h>
#include <linux/unistd.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h> 
#include <linux/compiler.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/moduleparam.h>
#include <linux/poll.h>
#include <linux/syscalls.h>
#include <linux/file.h>
#include "aiptouch.h"


MODULE_AUTHOR("IBM Retail Store Solutions");
MODULE_DESCRIPTION("IBM Touch Calibration RS-232 Read Monitor");
MODULE_LICENSE("GPL compatible");

/* Local Constants */
#define FALSE 0
#define TRUE  1

#define DEBUG
static const char DEVDRVR_ID[] = "aiptouch";
#define DEFAULT_MAJOR_DEV 250

#ifdef DEBUG
#define DPRINTK(fmt, args...) printk(KERN_DEBUG "%s(%s): " fmt, DEVDRVR_ID, __FUNCTION__ , ## args)
#else
#define DPRINTK(fmt, args...)
#endif


#define DRIVER_VERSION 1
#define BUFSIZE 1024

static int aiptouch_major = DEFAULT_MAJOR_DEV;
static DECLARE_WAIT_QUEUE_HEAD(aiptouch_queue);

void **sys_call_table = NULL; 


/* buffer data and spinlock thereof */
static int  bufpos_in = 0;
static int  bufpos_out = 0;
static char charbuffer[BUFSIZE];

asmlinkage ssize_t (*original_read) (unsigned int, char *, size_t);


typedef struct {
	int major;
	int minor;
	int hooked;
	int enabled;
	int dev_created;
} my_device;

static int aiptouch_disable_calibration(my_device *dev);

my_device touch_device;


/* aiptouch_sys_read: Intercepted read call
 * ************************************************************************
 *
 *  Function Name:  aiptouch_sys_read
 *
 *  Purpose:        Intercepted read call
 *
 *  Description:    For each sys_read call, check the major and minor to
 *                  determine if it is the device we are interested in
 *                  making a copy of the data.  Upon initial invocation,
 *                  call the original sys_read function to get the data.
 *
 *  Input:          fd  - file descriptor
 *                  buf - location in user address space tje data will be
 *                        stored.
 *                  count - number of bytes 
 *
 *  Output:         number of original bytes read.  
 *
 * ************************************************************************
 */

asmlinkage ssize_t aiptouch_sys_read(unsigned int fd, char *buf, size_t count)
{
	struct file *file = NULL;
	struct inode *inode = NULL;
	ssize_t rc = count;

	rc = original_read(fd, buf, count);

        if (rc == -1)
		goto out;

	if (touch_device.major == 0 && touch_device.minor == 0)
		goto out;

	if (fd >= NR_OPEN)
		goto out;

	file = current->files->fd[fd];
	if (!file)
		goto out;

	if (file->f_dentry == NULL)
		goto out;

	inode = file->f_dentry->d_inode;

	if (!inode)
		goto out;

	if (MAJOR(inode->i_rdev) == aiptouch_major)
		goto out;

	if ((touch_device.major != MAJOR(inode->i_rdev)) ||
	    (touch_device.minor != MINOR(inode->i_rdev)))
		goto out;

	// DPRINTK("Reading %d bytes for %d/%d\n", count, MAJOR(inode->i_rdev),
	// 	MINOR(inode->i_rdev));

	while (count > 0) {
		if (bufpos_in > BUFSIZE-1)
			bufpos_in = 0;

		get_user(charbuffer[bufpos_in++], (char *) buf++);
		count--;
	}

	wake_up_interruptible(&aiptouch_queue);

out:

	return rc;
}


/* aiptouch_open: Open device driver 
 * ************************************************************************
 *
 *  Function Name:  aiptouch_open
 *
 *  Purpose:        Open this device driver
 *
 *  Description:    
 *
 *  Input:          inode  -   
 *                  file   - 
 *
 *  Output:         0  -- success
 *
 * ************************************************************************
 */

static int aiptouch_open(struct inode *i, struct file *file)
{ 
	DPRINTK("hooked=%d, sct=0x%p\n", touch_device.hooked, sys_call_table);

	return 0; 
} 


/* aiptouch_close: Close the device handle
 * ************************************************************************
 *
 *  Function Name:  aiptouch_close
 *
 *  Purpose:        Close the device handle
 *
 *  Description:    When the application closes, make sure to restore the
 *                  sys_read function pointer in the sys_call_table.
 *
 *  Input:          inode  - not used
 *                  file   - not used
 *
 *  Output:         0  -- success
 *
 * ************************************************************************
 */

static int aiptouch_close(struct inode *i, struct file *file)
{ 
	aiptouch_disable_calibration(&touch_device);

	DPRINTK("hooked=%d, sct=0x%p\n", touch_device.hooked, sys_call_table);

	return 0;
} 


/* aiptouch_read: Read from device 
 * ************************************************************************
 *
 *  Function Name:  aiptouch_read
 *
 *  Purpose:        Read from this device
 *
 *  Description:    Read from the circular buffer
 *
 *  Input:          file  -   
 *                  buf   - 
 *                  count - 
 *                  ppos  - 
 *
 *  Output:         -EINVAL -- invalid
 *
 * ************************************************************************
 */

static int aiptouch_read(struct file *file, char *buf, size_t count,
			 loff_t *pos)
{
	int bytes_read = 0;

	if (bufpos_out == bufpos_in)
		goto out;

	if (count > BUFSIZE) {
		count = BUFSIZE;
	}

	while (count > 0) {
		put_user(charbuffer[bufpos_out++], buf++);
		count--;
		bytes_read++;

		if (bufpos_out > BUFSIZE-1)
			bufpos_out = 0;

		if (bufpos_out == bufpos_in)
			break;
	}

out:
	return bytes_read;
}


/* aiptouch_poll: Handler for polling
 * ************************************************************************
 *
 *  Function Name:  aiptouch_poll
 *
 *  Purpose:        Handler for polling
 *
 *  Description:    
 *
 *  Input:          file - file descriptor
 *                  wait - kernel poll list waiting structure
 *
 *  Output:         mask indicating if device has data for reading
 *
 * ************************************************************************
 */

static unsigned int aiptouch_poll(struct file *file,
				  struct poll_table_struct *wait)
{
	unsigned int mask = 0;

	poll_wait(file, &aiptouch_queue, wait);

	if (bufpos_in != bufpos_out) {
		mask = POLLIN | POLLRDNORM;
	}

	// DPRINTK("Poll Mask is 0x%x\n", mask);

	return mask;
}


/* aiptouch_enable_calibration: Get version information of this kernel file
 * ************************************************************************
 *
 *  Function Name:  aiptouch_enable_calibration
 *
 *  Purpose:        Hook the sys_read system call
 *
 *  Description:    Intercept calls from the /dev/tty device to allow
 *                  the calibration process read access to the touch
 *                  responses.  When X is running and using the tty 
 *                  device, it is unreliable to get the calibration
 *                  responses.
 *
 *  Input:          my_device - global data structure
 *                  buffer    - input parameters indicating what device
 *                              to intercept data from
 *
 *  Output:         0       - success
 *                  -ENOENT - calibration already enabled
 *                  -EFAULT - could not read request information
 *
 * ************************************************************************
 */

static int aiptouch_enable_calibration(my_device *dev, char *buffer)
{
	int   rc = 0;
	ENABLE_CALIBRATION_PARMS parms;

	DPRINTK("major=%d, minor=%d, hooked=%d, sct=0x%p\n",
		dev->major, dev->minor, dev->hooked, sys_call_table);

	if (dev->major != 0 || dev->enabled == FALSE ||
	    dev->hooked == TRUE) {
		rc = -ENOENT;
		goto out;
	}

	bufpos_in = 0;
	bufpos_out = 0;
	memset(charbuffer, 0, sizeof(charbuffer));

	/* Copy over the structure from user space */
	if (copy_from_user(&parms, buffer, sizeof(parms)) != 0) {
		rc = -EFAULT;
		goto out;
	}

	dev->major  = parms.major;
	dev->minor  = parms.minor;
	dev->hooked = TRUE;

	sys_call_table[__NR_read] = aiptouch_sys_read;

out:
	DPRINTK("rc=%d, hooked=%d, major=%d, minor=%d\n", rc, dev->hooked,
		dev->major, dev->minor);

	return rc;
}


/* aiptouch_disable_calibration: Stop intercepting read data
 * ************************************************************************
 *
 *  Function Name:  aiptouch_disable_calibration
 *
 *  Purpose:        Stop intercepting read data for a device
 *
 *  Description:    Restore the original sys_read function and stop
 *                  intercepting read data for a device.
 *
 *  Input:          buffer - pointer to argument structure
 *
 *  Output:         0       - success
 *
 * ************************************************************************
 */

static int aiptouch_disable_calibration(my_device *dev)
{
	int rc = 0;

	DPRINTK("major=%d, minor=%d, hooked=%d, sct=0x%p\n",
		dev->major, dev->minor, dev->hooked, sys_call_table);

	if (dev->enabled == TRUE && dev->hooked == TRUE) {
		dev->major  = 0;
		dev->minor  = 0;
		dev->hooked = FALSE;

		sys_call_table[__NR_read] = original_read;
	}

	DPRINTK("rc=%d\n", rc);

	return rc;
}


/* get_driver_version: Get version information of this kernel file
 * ************************************************************************
 *
 *  Function Name:  get_driver_version
 *
 *  Purpose:        Get version information of this kernel file
 *
 *  Description:    This function returns the current version information
 *                  if this kernel file.  Updated when new features are
 *                  added to allow user-mode application to adjust at
 *                  runtime to specific behaviors.
 *
 *  Input:          buffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - could not write to the user address space
 *
 * ************************************************************************
 */

static int aiptouch_driver_version(char *buffer)
{
	int   rc = 0;
	GET_DRIVER_VERSION_PARMS parms;

	DPRINTK("Driver version is %d\n", DRIVER_VERSION);

	parms.version    = DRIVER_VERSION;
	parms.reserved_1 = 0;
	parms.reserved_2 = 0;

	if (copy_to_user(buffer, &parms, sizeof(parms)) != 0)
		rc = -EFAULT;

	DPRINTK("rc=%d\n", rc);

	return rc;
}


/* aiptouch_clear_input: Clear all previously received data
 * ************************************************************************
 *
 *  Function Name:  aiptouch_clear_input
 *
 *  Purpose:        Clear all previously received data
 *
 *  Description:    Resets the input and output pointers to remove 
 *                  knowledge of any received data.
 *
 *  Input:          buffer - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EFAULT - could not write to the user address space
 *
 * ************************************************************************
 */

static int aiptouch_clear_input(void)
{
	bufpos_in = 0;
	bufpos_out = 0;

	return 0;
}


/* aiptouch_ioctl: I/O Control for this device driver 
 * ************************************************************************
 *
 *  Function Name:  aiptouch_ioctl
 *
 *  Purpose:        Handle ioctl calls to this device driver
 *
 *  Description:    Call correct function based on ioctl request.
 *
 *  Input:          inode  -   
 *                  file   - 
 *                  cmd    - ioctl request
 *                  arg    - pointer to argument structure
 *
 *  Output:         0       - success
 *                  -EINVAL - invalid request
 *                  -rc     - return value from called functions
 *
 * ************************************************************************
 */

static int aiptouch_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	int rc = 0;
	char *buffer = (char *) arg;

	switch (_IOC_NR(cmd)) {
	case _IOC_NR(ENABLE_CALIBRATION):
		rc = aiptouch_enable_calibration(&touch_device, buffer);
		break;
	case _IOC_NR(DISABLE_CALIBRATION):
		rc = aiptouch_disable_calibration(&touch_device);
		break;
	case _IOC_NR(GET_DRIVER_VERSION):
		rc = aiptouch_driver_version(buffer);
		break;
	case _IOC_NR(CLEAR_INPUT):
		rc = aiptouch_clear_input();
		break;
	default:
		DPRINTK("unrecognized command, cmd=0x%x\n", _IOC_NR(cmd));
		rc = -EINVAL;
		break;
	}

	DPRINTK("exiting, cmd=%d, rc=%d (0x%x)\n", _IOC_NR(cmd), rc, rc);

	return rc;
}


/* aiptouch_sys_call_table: Find the sys_call_table address
 * ************************************************************************
 *
 *  Function Name:  aiptouch_sys_call_table
 *
 *  Purpose:        Find the sys_call_table address
 *
 *  Description:    Find the location in the kernel of the sys_call_table.
 *                  We need the address in order to intercept the sys_read
 *                  calls.  The location in the sys_call_table we are
 *                  interested in is __NR_read.
 *
 *  Input:          None
 *
 *  Output:         TRUE  - found the address
 *                  FALSE - could not find the sys_call_table
 *
 * ************************************************************************
 */

static int aiptouch_sys_call_table(void)
{
	unsigned long *begin;
	int i;
	int rc = FALSE;

	begin = (unsigned long *) &system_utsname;

	for (i = 0; i < 1024; i++) {
		if (*(begin + i) == (unsigned long) sys_close) {
			sys_call_table = (void *) (begin + (i - __NR_close));

			if (sys_call_table[__NR_close] == (void *) sys_close) {
				DPRINTK("sys_call_table found at 0x%p\n",
					sys_call_table);

				rc = TRUE;
				break;
			}
		}
	}

	return rc;
}

/* The various file operations we support. */
static struct file_operations aiptouch_fops = {
	.read           = aiptouch_read,
	.poll           = aiptouch_poll,
	.ioctl          = aiptouch_ioctl,
	.open           = aiptouch_open,
	.release        = aiptouch_close,
	.owner          = THIS_MODULE,
};


/* aiptouch_init: Init device driver 
 * ************************************************************************
 *
 *  Function Name:  aiptouch_init
 *
 *  Purpose:        Initialize this device driver (called at insmod/modprobe
 *                  time)
 *
 *  Description:    Make a copy of the read system call from the
 *                  sys_call_table.
 *
 *  Input:          none
 *
 *  Output:         0      - success
 *                  -ENONT - Could not locate the sys_call_table
 *                  -rc    - Error return code from register_chardev
 *
 * ************************************************************************
 */

static int __init aiptouch_init(void)
{
	int rc = 0;

	/* Initialize our data structure */
	memset(&touch_device, 0, sizeof(touch_device));

	touch_device.major       = 0;
	touch_device.minor       = 0;
	touch_device.hooked      = FALSE;
	touch_device.enabled     = FALSE;
	touch_device.dev_created = FALSE;

	if (aiptouch_sys_call_table() == FALSE) {
		rc = -ENOENT;
		printk(KERN_ERR "%s: Failure to find sys_call_table, rc=%d\n",
		       DEVDRVR_ID, rc);
		goto out;
	}

	/* Register as a device with kernel. */
	if ((rc = register_chrdev(aiptouch_major, DEVDRVR_ID, &aiptouch_fops))) {
		rc = -rc;
		printk(KERN_ERR "%s: Failure to register device, rc=%d\n",
		       DEVDRVR_ID, rc);
		goto out;
	}

	original_read            = sys_call_table[__NR_read];
	touch_device.dev_created = TRUE;
	touch_device.enabled     = TRUE;

out:
	printk(KERN_INFO "%s: rc=%d, enabled=%d, created=%d, sct=0x%p, read=0x%p\n",
	       DEVDRVR_ID, rc, touch_device.enabled, touch_device.dev_created,
	       sys_call_table, original_read);

	return rc;
}


/* aiptouch_exit: Exit (unload) device driver 
 * ************************************************************************
 *
 *  Function Name:  aiptouch_exit
 *
 *  Purpose:        Unload this device driver (called at rmmod time)
 *
 *  Description:    Make sure the sys_read function in the sys_call_table
 *                  is restored to it's original value.
 *
 *  Input:          none
 *
 *  Output:         none
 *
 * ************************************************************************
 */

static void __exit aiptouch_exit(void)
{
	aiptouch_disable_calibration(&touch_device);

	if (touch_device.dev_created == TRUE) {
		if (unregister_chrdev(aiptouch_major, DEVDRVR_ID) != 0) {
			printk(KERN_ERR "%s: unregister_chrdev failed\n",
			       DEVDRVR_ID);
		}
	}

	printk(KERN_INFO "%s: Module unloaded\n", DEVDRVR_ID);
}

module_init(aiptouch_init);
module_exit(aiptouch_exit);

